-- Allows you to keep track of people you have had whisper conversations with

-- Variables
--TellTrack_WhisperMessages = { }
local TellTrack_WhisperMessages = { }
local TellTrack_OfflinePlayers = { }

-- hooked funtions
local SavedSendChatMessage = nil
local SavedFCF_SelectDockFrame = nil
local SavedChatFrame_OpenChat = nil
local SavedFCF_Close = nil
local SavedFCF_OpenNewWindow = nil

SELECTED_CHAT_FRAME = DEFAULT_CHAT_FRAME

-- executed on load, calls general set-up functions
function TellTrack:OnLoadCustom( frame )
	self:AddSlashCommand( nil, "/ttrack" )
	SlashCmdList["TELLTRACKRETELL"] = TellTrack_WhisperToPreviousTarget
	self:AddSlashCommand( "RETELL", "/retell", true )
	self:AddSlashCommand( "RETELL", "/re" )
end

function TellTrack:OnPlayerLogin( frame )
	-- these should definately not fire until setup is complete (after player_login)
	frame:RegisterEvent("PLAYER_ENTERING_WORLD")
	frame:RegisterEvent("CHAT_MSG_WHISPER")
	frame:RegisterEvent("CHAT_MSG_WHISPER_INFORM")
	frame:RegisterEvent("CHAT_MSG_SYSTEM")

	local db,o = self._db,self._options
	local _inst = self._acelib
	self.L = _inst.L
	self.LR = _inst.LR

	-- init vars
	o.Realm,o.Race,o.Player,o.Faction = (GetRealmName()),(UnitRace("player")),(UnitName("Player")),(UnitFactionGroup("Player"))
	o.RealmPostfix = "-" .. o.Realm
	o.RealmPostfixCmp = self:LowerUTF8(o.RealmPostfix)
	o.ListProfile = o.Realm .. " - " .. o.Faction
	o.PlayerProfile = o.Player .. " - " .. o.Realm

	-- curent conversation
	self._current = { name = "", showAll = true, }

	-- load frame list
	self._chat = db.Chat[o.PlayerProfile] or { }
	db.Chat[o.PlayerProfile] = self._chat

	-- load player list
	if o.ListSave ~= 1 then db.Log = { } end
	self._list = db.Log[o.ListProfile] or { }
	db.Log[o.ListProfile] = self._list

	-- update cmds and menus
	if _inst and _inst.menu then			-- remove default help option
		local args = _inst.menu.args
		local main = args.main; main.type = "toggle"; main.set = main.func;
			main.get = function() return TellTrackFrame:IsVisible() end
		args.options.passValue = "help"
		args.help.passValue = "info"
	end

	-- hook chat functions
	if SendChatMessage ~= SavedSendChatMessage then					-- Hook the chat event handler 'before'
		SavedSendChatMessage = SendChatMessage
		SendChatMessage = TellTrack_SendChatMessage
	end
	if FCF_SelectDockFrame ~= TellTrack_FCF_SelectDockFrame then	-- Hook dock frame selection 'after'
		SavedFCF_SelectDockFrame = FCF_SelectDockFrame
		FCF_SelectDockFrame = TellTrack_FCF_SelectDockFrame
	end
	if ChatFrame_OpenChat ~= TellTrack_ChatFrame_OpenChat then		-- Hook open chat to open as whisper
		SavedChatFrame_OpenChat = ChatFrame_OpenChat
		ChatFrame_OpenChat = TellTrack_ChatFrame_OpenChat
	end
	if FCF_Close ~= TellTrack_FCF_Close then						-- Hook open chat to open as whisper
		SavedFCF_Close = FCF_Close
		FCF_Close = TellTrack_FCF_Close
	end
	if FCF_OpenNewWindow ~= TellTrack_FCF_OpenNewWindow then		-- Hook open chat to open as whisper
		SavedFCF_OpenNewWindow = FCF_OpenNewWindow
		FCF_OpenNewWindow = TellTrack_FCF_OpenNewWindow
	end

	-- tweak ui
	TT_Button:OnAction(nil, "enable", o.Enable)
	TT_Button:OnSizeChanged( TellTrackFrame )
	TT_Button:UpdateBorderVisibility()
end


function TellTrack:OnCommandCustom( msg, cmd, args )
	local o,L = self._options, self.L
	--self:Debug("OnCommandCustom", msg, cmd, args)

	if not cmd then
		TT_Button:OnAction(nil, "enable")
		return true
	elseif TT_Button:OnAction(nil, cmd, args) then
		return true
	elseif cmd == "help" then
		self:Print(TT_CHAT_USEAGE)
		local list,short = TellTrack_cmdlist, self._meta.short
		for k,v in pairs(list) do
			self:Print("/"..short.." "..k.." - ".. L[v.help])
		end
	end
end

function TellTrack:Print_Help()
	self:Print(TT_CHAT_INFO)
	self:Print(TT_CHAT_INFO1)
	self:Print(TT_CHAT_INFO2)
	self:Print(TT_CHAT_INFO3)
end


function TellTrack_WhisperToPreviousTarget(msg)
	local o = TellTrack._options
	if not o.LastTell then return end
	SendChatMessage(msg, "WHISPER", o.LastLang, o.LastTell)
end


-- Handles events
function TellTrack:OnEventCustom(frame, event, ...)
	local ar1, ar2, ar3, ar4, ar5, ar6 = select(1, ...)
	if event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM" then
		self:HookAllAddMessages()
		if type(ar1) ~= "string" then return end

		--Special Hidden Whisper Ignores
		local len = ar1:len()
		if Guilded and len > 2 and ar1:sub(1,2) == "<G" and (len ~= 4 or ar1:sub(1,4) ~= "<GM>") then return	--Hide Guilded messages
		elseif GuildAds and len > 3 and ar1:sub(1,3) == "<GA" then return				--Hide GuildAds messages
		elseif len > 3 and ar1:sub(1,3) == "<T>" then return							--Hide Telepathy messages
		elseif len > 1 and ar1:sub(1,1) == "/" then return								--Hide AceComm messages
		elseif len > 7 and ar1:sub(1,7) == "SKP1019" then return						--Hide SKP messages
		end

		if event == "CHAT_MSG_WHISPER" then
			self:HandleMessageSentOrRecieved(ar2, false)
		end

		local name,cmpName = (self:NormName(ar2)), (self:NormNameCmp(ar2))
		if cmpName then
			TellTrack_OfflinePlayers[cmpName] = nil
		end

		local etype = string.sub(event,10)
		local message = gsub(ar1, "%%", "%%%%")

		self:Debug("TellTrack:OnEventCustom", "type",etype, "tag",ar6, "arg2",arg2, "name",name, "msg",message, "time",self:GetTime(nil, self._update.timediff))
		tinsert(TellTrack_WhisperMessages, {type=etype, tag=ar6, name=name, msg=message, time=self:GetTime(nil, self._update.timediff)})

	elseif event == "CHAT_MSG_SYSTEM" then
		if type(ar1) ~= "string" then return end

		local result = false;

		local _, _, name = strfind(ar1, TT_ERR_CHAT_PLAYER_NOT_FOUND_S)
		if name then result = true end

		if result == false then	_, _, name = strfind(ar1, TT_ERR_FRIEND_OFFLINE_S)
			if name then result = true end end

		if result == false then	_, _, name = strfind(ar1, TT_ERR_FRIEND_ONLINE_SS)
			if name then result = nil end end

		if name then
			TellTrack_OfflinePlayers[self:NormNameCmp(name)] = result
			self:UpdateTellTrackButtonsText()
		end
	end
end

-- adds a whisper to a frame
function TellTrack:GetTimeStamp(ts)
	local separator = "" -- ">"
	if ChatTimeStamps_TimeStamp and not ChatTimeStamps_SEPARATOR then
		separator = ""
	end
	if not ts then
		if ChatTimeStamps_LOCALTIME then
			ts = date()
		elseif Clock_GetTimeText then
			ts = Clock_GetTimeText()
		else
			ts = self:GetTime(nil, self._update.timediff)
		end
	end
	if type(ts) == "number" then
		ts = date("%H:%M:%S", ts)
	end
	if type(ts) ~= "string" then
		ts = ""
	end
	return ts..separator.." "
end

function TellTrack:RedrawWhisperFrame( )		--self:Debug("RedrawWhisperFrame", self._current and  self._current.name)
	local frame = self._chat.frame and _G[self._chat.frame]
	if not frame then return end
	local scrollOffset, found, _current = frame:GetCurrentScroll(), false, self._current
	frame:Clear()

	--Display only the valid whispers.
	local curNameCmp = self:NormNameCmp(_current.name)
	for i,v in pairs(TellTrack_WhisperMessages) do
		if _current.showAll or self:NormNameCmp(v.name) == curNameCmp then
			local info = ChatTypeInfo[v.type]
			local name, nameLocal = self:NormName(v.name)
			local part1 = _G["CHAT_"..v.type.."_GET"]..v.msg
			local part2 = (v.tag and _G["CHAT_FLAG_"..v.tag] or "").."|Hplayer:"..nameLocal.."|h".."["..nameLocal.."]".."|h"
			if TasteTheNaimbow_Loaded then
				part2 = "|cff"..TasteTheNaimbowExternalColor(nameLocal)..part2.."|r"
			end
			frame:AddMessage(format(part1, part2), info.r, info.g, info.b, info.id, nil, v.time)
			found = true
		end
	end

	frame:SetScrollOffset(scrollOffset)
	return found
end

function TellTrack:ShowConversationWith(name, showAll)		--self:Debug("ShowConversationWith", name, showAll)
	local db,o,_chat = self._db,self._options,self._chat
	if not _chat.frame or not _G[_chat.frame] then
		if o.AutoCreate==0 or not self:CreateWhisperChatFrame() then return	end
	end
	local frame = _chat.frame and _G[_chat.frame]
	local nameNorm,nameLocal = self:NormName(name)

	self._current = { name=nameNorm, showAll=showAll, }
	if not self:RedrawWhisperFrame() then
		self._current.showAll = true
		local timeStamp = o.TimeStamps
		o.TimeStamps = 0
		frame:AddMessage(format(self.L["-- No Current Conversation with %s --"],nameLocal))
		o.TimeStamps = timeStamp
	end
	_chat.name = nameNorm

	-- Show the frame and tab
	FCF_SetWindowName(frame, nameLocal)
	if frame.isDocked then
		FCF_SelectDockFrame(frame)
	end
end

function TellTrack:ShowMainChatFrame()
	local frame = DEFAULT_CHAT_FRAME
	-- Show the frame and tab
	if frame.isDocked then
		FCF_SelectDockFrame(frame)
	end
end

function TellTrack_ChatFrame_AddMessage(frame, text, r, g, b, id, addToStart, ts)
	local tt = TellTrack
	local AddMessage = SavedChatFrame_AddMessages[frame:GetName()]

	if frame:GetName() ~= tt._chat.frame then
		AddMessage(frame, text, r, g, b, id, addToStart)
		return					-- incorrect whisper frame
	end

	local stamps = ChatTimeStamps_ENABLED
	ChatTimeStamps_ENABLED = 0
	if strlen(text) < 5 then
		AddMessage(frame, text, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
	local timedText = text
	if tt._options.TimeStamps == 1 then
		timedText = tt:GetTimeStamp(ts)..text
	end
	if not tt._current or tt._current.showAll then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end

	local name = tt._current.name
	local nameNorm,nameLocal = tt:NormName(name)

	local namecolor = "ffffff"
	if TasteTheNaimbow_Loaded and TasteTheNaimbowExternalColor then
		namecolor = self:SafeCall(TasteTheNaimbowExternalColor, nameLocal)
	end
	--DEFAULT_CHAT_FRAME:AddMessage("Real "..string.gsub(string.gsub(text,":%d+\124","\124"),"\124","\\124"))
	--DEFAULT_CHAT_FRAME:AddMessage("Fake "..string.gsub(format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"),"\124","\\124"))
	--DEFAULT_CHAT_FRAME:AddMessage(tostring((tt:startsWith(string.gsub(text,":%d+\124","\124"), format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h")))))
	--local matchName = strlower(nameNorm:gsub("\-","\-"))
	local matchText = strlower(text:gsub(":%d+\124","\124"):gsub(":%d+:",":"))	-- remove numbers
		matchText = matchText:gsub(":whisper:[^\124^:]+","")					-- remove verbose command

	local Hplayer = "|Hplayer:"..nameNorm.."|h["..nameNorm.."]|h"
	local HplayerL = "[|Hplayer:"..nameNorm.."|h"..nameNorm.."|h]"
	local HplayerLocal = "|Hplayer:"..nameLocal.."|h["..nameLocal.."]|h"
	local HplayerLocalL = "[|Hplayer:"..nameLocal.."|h"..nameLocal.."|h]"
	local HplayerMixed = "|Hplayer:"..nameNorm.."|h["..nameLocal.."]|h"
	local HplayerMixedL = "[|Hplayer:"..nameNorm.."|h"..nameLocal.."|h]"
	local function nameMatch(get, name)
		local s, s2 = matchText, strlower(format(get, name));
		if type(s) ~= "string" or type(s2) ~= "string" then return false end
		return s == s2 or (s:len() > s2:len() and string.sub(s, 1, s2:len()) == s2)
	end
	if	-- this whole section, above and below, needs much love, not most efficent.  need early outs, etc...
		nameMatch(CHAT_WHISPER_GET, Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..Hplayer) or
		nameMatch(CHAT_WHISPER_INFORM_GET, Hplayer) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerL) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerL) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerL.."|r") or
		
		nameMatch(CHAT_WHISPER_GET, HplayerLocal) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerLocal) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerLocal) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerLocal) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerLocal) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerLocal.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerLocal.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerLocal.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerLocal.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerLocal.."|r") or
		nameMatch(CHAT_WHISPER_GET, HplayerLocalL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerLocalL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerLocalL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerLocalL) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerLocalL) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerLocalL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerLocalL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerLocalL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerLocalL.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerLocalL.."|r") or
		
		nameMatch(CHAT_WHISPER_GET, HplayerMixed) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerMixed) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerMixed) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerMixed) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerMixed) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerMixed.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerMixed.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerMixed.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerMixed.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerMixed.."|r") or
		nameMatch(CHAT_WHISPER_GET, HplayerMixedL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerMixedL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerMixedL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerMixedL) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerMixedL) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerMixedL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerMixedL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerMixedL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerMixedL.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerMixedL.."|r")
	then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
end


function TellTrack:HookAllAddMessages()
	if SavedChatFrame_AddMessages then return end

	SavedChatFrame_AddMessages = {}
	for i=1, NUM_CHAT_WINDOWS do
		local frameName = "ChatFrame"..i
		local frame = _G[frameName]
		if frame and (frame.AddMessage ~= TellTrack_ChatFrame_AddMessage) then
			SavedChatFrame_AddMessages[frameName] = frame.AddMessage
			frame.AddMessage = TellTrack_ChatFrame_AddMessage
		end
	end
end

-- finds 
function TellTrack:FindChatFrame( ... )
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local chatFrame = _G[sChatFrame]
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		local sNameCmp = self:NormNameCmp(sName)
		if select("#",...) == 0 then
			if (not bShown and not chatFrame.isDocked) or i == NUM_CHAT_WINDOWS then
				return chatFrame
			end
		elseif bShown or chatFrame.isDocked then
			for i1 = 1, select("#",...) do
				local name = select(i1,...)
				if name and self:NormNameCmp(name) == sNameCmp then
					return sChatFrame,sName
				end
			end
		end
	end
end

function TellTrack:CreateWhisperChatFrame()
	local _chat = self._chat
	if _chat.frame and _G[_chat.frame] then return _chat.frame end

	local defaultName, chatName, name = self.L["Tell Track"], _chat.name
	local chatNameCmp,defaultNameCmp = self:NormNameCmp(defaultName),self:NormNameCmp(chatName)

	_chat.frame,name = self:FindChatFrame(chatName, chatNameCmp~=defaultNameCmp and defaultName);	-- get if already exists
	if not _chat.frame then
		if FCF_GetNumActiveChatFrames() >= NUM_CHAT_WINDOWS then return end				-- dont get if have to take someone elses
		name = chatName or defaultName
		local selected = SELECTED_DOCK_FRAME
		FCF_OpenNewWindow(select(2,self:NormName(name)))
		FCF_SelectDockFrame(selected)
		_chat.frame = self:FindChatFrame(name)
	end
	if _chat.frame then																	-- make sure clear out
		_chat.name = self:NormName(name)
		local chatFrame = _G[_chat.frame]
		ChatFrame_RemoveAllMessageGroups(chatFrame)
		ChatFrame_AddMessageGroup(chatFrame, "WHISPER")
		return true
	end
end

function TellTrack:HideWhispersInOtherFrames()
	local o, sframe = self._options, self._chat.frame
	local frame = sframe and _G[sframe]
	if not frame then return end
	local func = ChatFrame_RemoveMessageGroup -- (o.HideWhispers == 1) and ChatFrame_RemoveMessageGroup or ChatFrame_AddMessageGroup
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local chatFrame = _G[sChatFrame]
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		if chatFrame and (bShown or chatFrame.isDocked) and sChatFrame ~= sframe then
			func(chatFrame, "WHISPER")
		end
	end
end


function TellTrack_SendChatMessage(text, type, language, target)
	TellTrack:SendChatMessage(text, type, language, target)
end

-- why not using _INFORM event? would it show out of order?
function TellTrack:SendChatMessage(text, type, language, target)
	local o = self._options
	if type == "WHISPER" then
		o.LastTell = target
		o.LastLang = language
		if o.Enable == 1 and strsub(text, 1, 1) ~= "<" then	-- prevent data message transfers from being used
			self:HandleMessageSentOrRecieved(target, true)
		end
	end
	if SavedSendChatMessage then
		SavedSendChatMessage(text, type, language, target)
	end
end

function TellTrack_ChatFrame_OpenChat(text, chatFrame)
	local tt = TellTrack
	local _chat, _current = tt._chat, tt._current
	if text == "" and _current and _current.name and _current.name ~= ""
			and SELECTED_CHAT_FRAME:GetName() == _chat.frame then
		tt:InitiateWhisperToTarget(_current.name)
	else
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat(text, chatFrame)
		end
	end
end

function TellTrack_FCF_OpenNewWindow( name )
	if not SavedFCF_OpenNewWindow then return end
	SavedFCF_OpenNewWindow( name )
	local chatFrame = (TellTrack._options.HideWhispers == 1) and TellTrack:FindChatFrame() or nil
	if chatFrame then
		ChatFrame_RemoveMessageGroup(chatFrame, "WHISPER")
	end
end

function TellTrack_FCF_Close( chatFrame, fallBack )
	local _chat = TellTrack._chat
	local frame = fallBack or chatFrame or FCF_GetCurrentChatFrame()
	if frame and _chat and frame:GetName() == _chat.frame then
		_chat.frame = nil
	end
	if SavedFCF_Close then
		SavedFCF_Close(chatFrame, fallBack)
	end
end

function TellTrack_FCF_SelectDockFrame(frame)
	local tt = TellTrack
	local o,list = tt._options,tt._list

	if SavedFCF_SelectDockFrame then
		SavedFCF_SelectDockFrame(frame)
	end
	if frame:GetName() == tt._chat.frame then
		if tt._current.showAll then
			for i = 1, o.ListSize do
				if list[i] then	list[i].unread = 0 end
			end
		elseif tt._current.name ~= "" then
			local curNameCmp = tt:NormNameCmp(tt._current.name)
			for i = 1, o.ListSize do
				if list[i] and tt:NormNameCmp(list[i].name) == curNameCmp then
					list[i].unread = 0
				end
			end
		end
		tt:UpdateTellTrackButtonsText()
	end
	SELECTED_CHAT_FRAME = frame;		--FIXME: is necessary
end

function TellTrack:InitiateWhisperToTarget(name)
	self:InitWhisper(select(2,self:NormName(name)), nil, SavedChatFrame_OpenChat)
end

function TellTrack:UpdateTellTrackButtonsText()
	local list = self._list
	if not list then return end
	local o = self._options
	local maxitems = TellTrackFrame.ButtonCount
	local listcount = #list
	local lastToFmtStr = "|cff33ff33%s|r"
	local lastFromFmtStr = "|cffff3333%s|r"
	local noNameFmtStr = "|cff666666%s|r"

	for i=1, maxitems do
		if o.ListInvert ~= 1 then
			id = (i + o.ListOffset) - 1
		else
			id = (o.ListSize - (i + o.ListOffset)) + 2
		end

		local item = list[id]
		local sline = "TellTrack" .. i
		local line = _G[sline]

		if not line then
		else
			local fmt, val = noNameFmtStr, "Empty"
			local lineText = _G[sline.."Text"]
			if not item or not item.dispName then
				line.playerid = nil
				line.item = nil
				line.title = nil
			else
				line.title = item.dispName
				--line.title = select(2, self:NormName(item.dispName))
				line.playerid = id
				line.item = item
				val = select(2, self:NormName(item.dispName))
				if not TellTrack_OfflinePlayers[self:NormNameCmp(item.dispName)] then
					fmt = (item.sentTo) and lastToFmtStr or lastFromFmtStr
				end
				if item.unread and item.unread > 0 then
					val = "("..item.unread..") "..val
				end
				if o.LineNums == 1 then
					val = id .. ". "..val
				end
			end
			lineText:SetText(format(fmt, val))
			lineText:Show()
		end
	end
end

function TellTrack:CapUTF8( s )
	if not s then return end
	local s1,len,c,c2 = "",s:len(),32
	for i = 1, len do
		c2 = c; c = s:byte(i);
		if c<=127 and (c2 == 45 or c2 == 32) then s1=s1..strupper(strchar(c));
		elseif c<=127 then	s1=s1..strlower(strchar(c));
		else				s1=s1..strchar(c); end
	end
	return s1
end

function TellTrack:LowerUTF8( s )
	if not s then return end
	return strlower(s)	--fixme: not really utf-8 safe, re-check all the strlower's later
end

function TellTrack:NormName( s )
	if not s then return end
	local o,disp = self._options, self:CapUTF8(s)
	local dispR = disp:gsub(o.RealmPostfix.."$","")
	if not disp:find("-") then disp = dispR .. o.RealmPostfix end
	return disp, dispR
end

function TellTrack:NormNameCmp( s )
	if not s then return end
	local o,cmp = self._options, self:LowerUTF8(s)
	local cmpR = cmp:gsub(o.RealmPostfixCmp.."$","")
	if not cmp:find("-") then cmp = cmpR .. o.RealmPostfixCmp end
	return cmp, cmpR
end

function TellTrack:HandleMessageSentOrRecieved( target, isSent )
	if not target then return end
	local o = self._options
	local max,_current = o.ListSize,self._current

	local normName,localName = self:NormName(target)
	local cmpName,cmpLName = self:NormNameCmp(target)

	-- find
	local item, id = nil, 0
	for k, v in pairs(self._list) do
		if v and v.cmpName == cmpName then item,v = v,nil; end
		if v then id = id + 1; if id ~= k then self._list[id] = v; end end		-- compress as we go
	end
	while (id >= max) do tremove(self._list, 1); id = id - 1; end
	if not item then item = { target = target, name = normName, cmpName = cmpName, dispName = normName, unread = 0, } end
	id = id + 1; self._list[id] = item

	-- update
	item.sentTo = isSent
	local chatFrameName = self._chat.frame
	if not item.unread then	item.unread = 0	end
	if not isSent and chatFrameName then
		if not _G[chatFrameName]:IsVisible() or (self:NormNameCmp(_current.name) ~= cmpName and not _current.showAll) then
			item.unread = item.unread + 1
		end
	end

	-- change offset
	if o.ListInvert == 1 then
		if id < max - TellTrackFrame.ButtonCount and (max - id + 1) ~= o.ListOffset then
			self:ChangeArrayOffset(max-id+1)
		end
	else
		if id > TellTrackFrame.ButtonCount and (id-TellTrackFrame.ButtonCount+1) ~= o.ListOffset then
			self:ChangeArrayOffset(id-TellTrackFrame.ButtonCount+1)
		end
	end

	self:UpdateTellTrackButtonsText()
end

