﻿
TT_Button = LibKarmaUI:GetInstance( TellTrack, {
	addItemInfo = { func = nil, value = { }, },
	funcMenuOnClick	= function(frame) return TT_Button:Menu_OnClick(frame) end,
	funcMenuOnGet	= function(frame) return TT_Button:Menu_OnGet(frame) end,
	funcMenuOnSet	= function(...) return TT_Button:Menu_OnSet(...) end,
})

function TT_Button:ChangeValue(name, v)
	local tt = TellTrack
	local o,L = tt._options, tt.L

	local _cmd = TellTrack_cmdlist[name]
	if not _cmd or _cmd.type ~= "toggle" then return end

	local oname = _cmd and _cmd.var
	if not oname then return end
	local old = o[oname]
	v = (v=="on" or v=="1" or v==1) and 1 or ((v=="off" or v=="0" or v==0) and 0 or ((old==1) and 0 or 1))

	if old ~= v then
		o[oname] = v
		if oname~="Enable" and _cmd[v] then tt:Print(L[_cmd[v]]) end
		return true
	end
	return false
end

function TT_Button:OnAction( frame, cmd, value )
	local tt = TellTrack
	local o = tt._options

	cmd = cmd or (frame and frame._type); value = value or (frame and frame._value)

	local changed = self:ChangeValue(cmd, value)

	if cmd == "enable" then
		if o.Enable == 0 then
			TellTrackFrame:Hide()
		else
			TellTrackFrame:Show()
		end
	elseif cmd == "info" then
		tt:Print_Help()
	elseif cmd == "invert" then
		if changed then
			tt:ChangeArrayOffset((o.ListInvert == 1) and o.ListSize or 1)
			tt:UpdateTellTrackButtonsText()
		end
	elseif cmd == "hidewhispers" then
		if o.HideWhispers == 1 then
			tt:HideWhispersInOtherFrames()
		end
	elseif cmd == "timestamps" then
		tt:RedrawWhisperFrame()
	elseif cmd == "clearall" then
		local db = tt._db
		tt._list = { }; db.Log[o.ListProfile] = tt._list
		tt:UpdateTellTrackButtonsText()
	elseif cmd == "border" then
		self:UpdateBorderVisibility()
	elseif cmd == "linenums" then
		tt:UpdateTellTrackButtonsText()
	elseif cmd == "showall" then
		tt:ShowConversationWith(tt._current.name, true)
	elseif cmd == "createcf" then
		tt:CreateWhisperChatFrame()
	elseif cmd == "whisper" then
		local item = value and tt._list[value]
		if not item then return false end
		tt:InitiateWhisperToTarget(item.name)
	elseif cmd == "whisperlast" then
		if not o.LastTell then return false end
		tt:InitiateWhisperToTarget(o.LastTell)
	elseif cmd == "showconv" then
		local item = value and tt._list[value]
		if not item then return false end
		tt:ShowConversationWith(item.name, false)
	elseif cmd == "who" then
		local item = value and tt._list[value]
		if not item then return false end
		SendWho("n-\""..item.dispName.."\"")
		tt:ShowMainChatFrame()
	elseif cmd == "invite" then
		local item = value and tt._list[value]
		if not item then return false end
		InviteUnit(item.name)
	elseif cmd == "addfriend" then
		local item = value and tt._list[value]
		if not item then return false end
		AddFriend(item.name)
	elseif cmd == "delete" then
		local item = value and tt._list[value]
		if not item then return false end
		tremove(tt._list, value)
		tt:UpdateTellTrackButtonsText()
	elseif TellTrack_cmdlist[cmd] ~= nil then
		return true									-- handled even if no code
	else
		return false								-- not handled
	end

	return true										-- handled in the list
end


-- ************** menus **************

function TT_Button:Menu_OnClick( frame )
	if not frame then return end
	local pass,pass2 = frame.arg1, frame.arg2
	if not pass then return end

	if pass.frameName then
--		RT_Dialog:Show(v.frameName, v.type, v.what, v.raidid, v.itemid, v.playerid, v.zone, v.boss)
	else
		local parent = frame:GetParent(); parent = parent and parent.dropdown:GetParent()
		self:OnAction(nil, pass, (pass2 and parent) and parent[pass2])
	end
end

function TT_Button:Menu_OnSet( frame )
	local pass = frame and frame.arg1
	if pass then
		self:OnAction(nil, pass, nil)
	end
end

function TT_Button:Menu_OnGet( frame )
	local tt = TellTrack
	local o = tt._options
	local pass = frame and frame.arg1
	if pass then
		local cmd = TellTrack_cmdlist[pass]
		return o[cmd.var] == 1
	end
end

function TT_LineTemplate_RightClickMenu_Initialize( frame, level )
	if not level then return end
	local self = TT_Button
	local inst = self._karma
	local L = inst.L
	local parent = frame._parent or frame:GetParent()

	if not self._rcmenu2 then
		local mf,gf,sf = self.funcMenuOnClick, self.funcMenuOnGet, self.funcMenuOnSet
		self._rcmenu2 = { handler = self, type = "group", sort=0, args = {
			{ name = L["Tell Track"], isTitle = 1, },
			{ pass="whisper", type="execute", name=L[TT_MENU_WHISPER], func=mf, pass2="playerid", },
			{ pass="showconv",type="execute", name=L[TT_MENU_SHOWCONV], func=mf, pass2="playerid", },
			{ pass="who", type="execute", name=L[TT_MENU_WHO], func=mf, pass2="playerid", },
			{ pass="invite", type="execute", name=L[TT_MENU_GRPINV], func=mf, pass2="playerid", },
			{ pass="addfriend", type="execute", name=L[TT_MENU_ADDFRIEND], func=mf, pass2="playerid", },
			{ pass="delete", type="execute", name=L[TT_MENU_DELETE], func=mf, pass2="playerid", },
			{ name="--------------", isTitle = 1, },
			{ pass="clearall", type="execute", name=L[TT_MENU_DELETE_ALL], func=mf, },
			{ pass="invert", type="toggle",  name=L[TT_MENU_INVERT], get=gf, set=sf, },
			{ pass="showall", type="execute", name=L[TT_MENU_SHOWWHISPS], func=mf, },
			{ pass="createf", type="execute", name=L[TT_MENU_CREATEFRAME], func=mf, },
			{ pass="border", type="toggle",  name=L[TT_MENU_BORDER], get=gf, set=sf, },
			{ pass="linenums", type="toggle",  name=L[TT_MENU_LINENUMS], get=gf, set=sf, },
		}, }
	end
	self:GenerateMiniMenu(parent, self._rcmenu2)
end


-- ************** buttons **************

function TT_Button:OnClick( frame, button )
	local parent = frame._parent or frame:GetParent()
	if self._super.OnClick( self, frame, button ) then return true end

	local o = self._karma._options

	if frame.type == "player" then
		if button == "RightButton" then
			TT_LineTemplate_RightClickMenu_Initialize(frame, 0)
		else
			TellTrack:TextButton_OnClick(frame, button)
		end
	elseif frame.type == "arrow" then
		local isUp = (frame.what == "up")
		if button == "RightButton" then
			TellTrack:ChangeArrayOffset( isUp and 1 or o.ListSize )
		elseif button == "LeftButton" then
			TellTrack:ChangeArrayOffset( o.ListOffset + (TellTrackFrame.ButtonCount * (isUp and -1 or 1)) )
		else
			return
		end
		TellTrack:UpdateTellTrackButtonsText()
	elseif frame.type == "qbutton" then
		self:OnAction(frame, "info")
	end
end

function TellTrack:TextButton_OnClick(frame, button)
--	TellTrack:DebugTable("TT_Button:OnEnter", frame:GetName(), frame )
	if not frame.item then return end
	if button == "LeftButton" then
		if (IsShiftKeyDown() and IsAltKeyDown()) then
			TT_Button:OnAction(frame, "delete", frame.playerid)
		elseif (IsShiftKeyDown()) then
			TT_Button:OnAction(frame, "who", frame.playerid)
		elseif (IsAltKeyDown()) then
			TT_Button:OnAction(frame, "invite", frame.playerid)
		elseif (IsControlKeyDown()) then
			TT_Button:OnAction(frame, "addfriend", frame.playerid)
		else
			local name = frame.item.name
			local eb = SELECTED_DOCK_FRAME.editBox
			if self._options.WhisperFirst == 1 then
				if (eb:IsVisible()) and (eb:GetAttribute("chatType") == "WHISPER") and
						(self:NormNameCmp(eb:GetAttribute("tellTarget")) == self:NormNameCmp(name)) then
					self:ShowConversationWith(name, false)
				else
					self:InitiateWhisperToTarget(name)
				end
			else
				if (SELECTED_DOCK_FRAME == _G[self._chat.frame]) and
						(self:NormNameCmp(self._current.name) == self:NormNameCmp(name)) then
					self:InitiateWhisperToTarget(name)
				else
					self:ShowConversationWith(name, false)
				end
			end
		end
	end
end


function TT_Button:OnEnter( frame )
	local parent = frame._parent or frame:GetParent()
	local title, text, f
--	TellTrack:DebugTable("TT_Button:OnEnter", frame:GetName(), frame )

--	TellTrack:SetBorderVisibility( 1 )
--	f = _G[parent:GetParent():GetName().."Tab"]
--	if f then f:Show() end

	self._super.OnEnter(self, frame, title, text, true )
end


function TT_Button:OnLeave( frame )
	parent = parent or frame:GetParent()
	local f

--	TellTrack:SetBorderVisibility( 0 )
--	f = _G[parent:GetParent():GetName().."Tab"]
--	if f then f:Hide() end

	self._super.OnLeave(self, frame)
end

function TellTrack:ChangeArrayOffset(offset)
	local o = self._options
--	self:Debug("ChangeArrayOffset",offset)
	if offset <= 0 then												-- lower bounds
		offset = 1
	end
	local capSize = o.ListSize - TellTrackFrame.ButtonCount + 1		-- upper bounds
	if offset > capSize then
		offset = capSize
	end
	o.ListOffset = offset
end

function TT_Button:OnMouseWheel(frame, v)
	local tt = TellTrack
	local o,list = tt._options,tt._list

	local v = v < 0 and 1 or (v > 0 and -1 or 0)
	tt:ChangeArrayOffset(o.ListOffset -  v)
	tt:UpdateTellTrackButtonsText()
end

function TT_Button:OnSizeChanged( frame )
	local parent = frame._parent or frame:GetParent()
	local fname = frame and frame:GetName()

	local tt = TellTrack
	local o,list = tt._options,tt._list
	if not fname or not list then return end				-- onsizechange can fire before tt is initialized

	if frame.type == "mainframe" then
		local height = frame:GetHeight() - 44				-- insets and arrow buttons
		local count = floor(height / 20)
		frame.ButtonCount = frame.ButtonCount or 0

		if count ~= frame.ButtonCount then
			local max, id = o.ListSize, #list
			local oh = frame.ButtonCount - id; oh = (oh < 0) and 0 or oh;
			if o.ListInvert ~= 1 then
				o.ListOffset = (o.ListOffset + (frame.ButtonCount - count)) - oh;
			else
				o.ListOffset = id + oh
			end
			tt:ChangeArrayOffset(o.ListOffset)

			frame.ButtonCount = count
			for i=1, o.ListSize do
				local sline = "TellTrack"..i
				local line = _G[sline]
				if i > frame.ButtonCount then line:Hide() else line:Show() end
			end
			tt:UpdateTellTrackButtonsText()
		end

		local offset = floor((height - (count * 20) + 20) / 2)
		TellTrackFrameArrowUpButton:SetHeight(offset)
		TellTrackFrameArrowDownButton:SetHeight(offset)
	end
end


function TT_Button:UpdateBorderVisibility()
	if TellTrack._options.Border == 1 then
		TellTrackFrameBackground:Hide()
		TellTrackFrameArrowUpButton:Hide()
		TellTrackFrameArrowDownButton:Hide()
		TellTrackFrame:SetHitRectInsets(10, 10, 10+TellTrackFrameArrowUpButton:GetHeight(), 10+TellTrackFrameArrowDownButton:GetHeight())
	else
		TellTrackFrameBackground:Show()
		TellTrackFrameArrowUpButton:Show()
		TellTrackFrameArrowDownButton:Show()
		TellTrackFrame:SetHitRectInsets(0, 0, 0, 0)
	end
end
